## Lab session: segmentation by dynamic programming

set.seed(1)
y <- rnorm(10, sd=0.2)
y[3:5] <- y[3:5] + 1

getJnaive <- function(y) {
    n <- length(y)
    J <- matrix(NA, ncol=n, nrow=n)
    for (ii in 1:n) {
        for(jj in ii:n) {
            avg <- mean(y[ii:jj])
            diffs <- y[ii:jj]-avg
            J[ii,jj] <- sum(diffs^2)
        } ## for (jj ...
    } ## for (ii ...
    return(J)
}

getJ <- function(y) {
    n <- length(y)
    S <- c(0, cumsum(y))
    T <- c(0, cumsum(y^2))
    J <- matrix(NA, ncol=n, nrow=n)
    for (ii in 1:n) {
        for(jj in ii:n) {
            J[ii, jj] <- T[jj+1]-T[ii] - (S[jj+1]-S[ii])^2/(jj-ii+1)
        } ## for (jj ...
    } ## for (ii ...
    return(J)
}

system.time(res <- getJ(rnorm(2000)))

y <- rnorm(100)
resN <- getJnaive(y)
res <- getJ(y)
stopifnot(all.equal(res, resN))

library("rbenchmark")
y <- rnorm(5e2)
benchmark(getJ(y), getJnaive(y), replications=10)[, 1:4]

# check that getJ is quadratic in n
benchmark(getJ(rnorm(1e2)), getJ(rnorm(1e3)), getJ(rnorm(2e3)), replications=10)[, 1:4]



## step 2: induction

## Dynamic programming
V <- matrix(numeric((K+1)*n), ncol=n)
## V[i,j] is the best RSE for segmenting intervals 1 to j
## with at most i-1 change points
bkp <-   matrix(integer(K*n), ncol=n)
## bkp[i, j] is the *last* bkp of the best segmentation of [1:j] in i segments

## With no change points (i=1), V[i,j] is just the precomputed RSE
## for intervals 1 to j
V[1, ] <- J[1, ]

## Then we apply the recursive formula
KK <- seq(length=K)
for (ki in KK){
    for (jj in (ki+seq(length=n-ki))){
        idxh <- ki:(jj-1)
        obj <- V[ki, idxh] + J[idxh+1, jj]
        val <- min(obj)
        ind <- which.min(obj)
        V[ki+1, jj] <- val
        bkp[ki, jj] <- ind + ki-1L
    }
}

## TO DO: backtracking